#!/usr/bin/env python
#
# trace.py - debugging functions
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides some useful logging/debugging functions.

.. warning:: This module is not intended for general use - it is solely for
             development/debugging purposes. Do not use it unless you know
             know what you are doing.


.. warning:: When this module is imported, it monkey patches the
             :class:`.CallQueue` class in a devious and dangerous manner to
             allow for more informative debug statements.  Therefore it's a
             bad idea to even import this module, unless you really know what
             you are doing.


.. warning:: Just don't import this module, ok?


The functions provided by this module are as follows:

.. autosummary::
   trace
   setcause
   propchange
"""

import            logging
import            inspect
import os.path as op


log = logging.getLogger(__name__)


# If trace debugging is enabled, we're going to
# do some funky stuff to the props callqueue
# object, so we can get some extra information
# in the propchange function.
if log.getEffectiveLevel() == logging.DEBUG:

    log.debug('Monkey-patching fsleyes_props.properties_value.queue instance')

    import fsleyes_props as props

    import six.moves.queue as queue

    # The problem that I am addressing here is the
    # fact that, when a property value listener is
    # called, the cause of the call (i.e. the point
    # at which the property value was changed) does
    # not necessarily exist in the stack. This is
    # because when a PV instance passes all of its
    # listeners to the CallQueue (CQ) to call, the
    # CQ will queue them immediately, but will not
    # necessarily call them - they are only called
    # if the CQ is not already processing the
    # listeners from another PV change.

    # A single CallQueue instance is shared
    # by all PropertyValue instances to queue/
    # call property value listeners.
    theQ = props.properties_value.PropertyValue.queue


    def tracePop(*args, **kwargs):

        # When the CallQueue calls its _pop
        # method, it will call the returned
        # function. For debugging purposes,
        # we'll pop the associated cause
        # (enqueued in the tracePush function
        # below), so the propchange function
        # can print it out.
        #
        # This will throw Queue.EmptyError if
        # the _causes queue is empty, but
        # that is what the real _pop method
        # does anyway.
        try:    theQ._currentCause = theQ._causes.get_nowait()
        except: theQ._currentCause = None

        return theQ._realPop(*args, **kwargs)


    def tracePush(*args, **kwargs):

        pushed = theQ._realPush(*args, **kwargs)

        # If the real _push method returns False,
        # it means that the function was not enqueued
        if not pushed:
            return False

        frames = inspect.stack()[1:]

        # We search for the first frame in the stack
        # which is not in the props package  - this
        # will be the point of the PV change which
        # caused this listener to be enqueued
        triggerFrame = None
        for frame in frames:
            if 'fsleyes_props' not in frame[1]:
                triggerFrame = frame
                break

        # This should never happen,
        # but in case it does, we
        # put a dummy value into
        # the causes queue, so it
        # is the same length as the
        # reall call queue
        if triggerFrame is None:
            theQ._causes.put_nowait(None)

        # Store the cause of the listener
        # push on the causes queue
        else:

            cause = [triggerFrame[1],
                     triggerFrame[2],
                     triggerFrame[3]]
            if triggerFrame[4] is not None:
                cause.append(triggerFrame[4][triggerFrame[5]])
            else:
                cause.append('<input>')
            theQ._causes.put_nowait(cause)

        return True

    # Patch the CallQueue instance with
    # our push/pop implementations
    theQ._causes          = queue.Queue()
    theQ._realPush        = theQ._CallQueue__push
    theQ._realPop         = theQ._CallQueue__pop
    theQ._CallQueue__push = tracePush
    theQ._CallQueue__pop  = tracePop


def trace(desc):
    """Outputs a log message containing the given description and the current
    stack trace.
    """

    if log.getEffectiveLevel() != logging.DEBUG:
        return

    stack = inspect.stack()[1:]
    lines = '{}\n'.format(desc)

    for i, frame in enumerate(stack):

        srcMod    = frame[1]
        srcLineNo = frame[2]

        if frame[4] is not None: srcLine = frame[4][frame[5]]
        else:                    srcLine = '<native>'

        lines = lines + '{}{}:{}: {}\n'.format(
            ' ' * (i + 1),
            srcMod, srcLineNo,
            srcLine.strip())

    log.debug(lines)

    return lines


def propchange(*args):
    """Intended to be called from a :class:`.PropertyValue` listener
    function.

    If this function is called due to a change, attempts to determine the line
    of code which triggered the change. Prints some informative log messages.

    :arg args: The arguments that were passed to the listener function.
    """


    if log.getEffectiveLevel() != logging.DEBUG:
        return

    import fsleyes_props as props

    theQ  = props.properties_value.PropertyValue.queue
    stack = inspect.stack()

    listenerFile = stack[1][1]
    listenerLine = stack[1][2]
    listenerFunc = stack[1][3]

    triggerFile = None

    if len(args) != 4:
        triggerFile = stack[2][1]
        triggerLine = stack[2][2]
        triggerFunc = stack[2][3]
        triggerSrc  = stack[2][4][stack[2][5]]
    else:
        triggerFile = theQ._currentCause[0]
        triggerLine = theQ._currentCause[1]
        triggerFunc = theQ._currentCause[2]
        triggerSrc  = theQ._currentCause[3]

    if triggerFile is None:
        log.debug('Listener {} ({}:{}) was called '
                  'due to an unknown process'.format(
                      listenerFunc,
                      op.basename(listenerFile),
                      listenerLine))

    else:

        if len(args) != 4:
            reason = 'manually called from'
        else:
            value, valid, ctx, name = args
            reason = 'called due to a value change of {}.{} ({}) at'.format(
                type(ctx).__name__,
                name,
                value)

        log.debug('Listener {} ({}:{}) was {} '
                  '{} ({}:{}:{})'.format(
                      listenerFunc,
                      op.basename(listenerFile),
                      listenerLine,
                      reason,
                      triggerFunc,
                      op.basename(triggerFile),
                      triggerLine,
                      triggerSrc.strip()))


def setcause(desc):
    """I can't quite remember the difference betwen this function and the
    :func:`propchange` function.
    """

    if log.getEffectiveLevel() != logging.DEBUG:
        return

    stack      = inspect.stack()[1:]
    causeFrame = None
    ultCauseFrame = None

    for i, frame in enumerate(stack):

        if 'fsleyes_props' not in frame[1]:
            causeFrame = frame
            break

    if causeFrame is not None:
        for i, frame in reversed(list(enumerate(stack))):
            if 'fsleyes_props' in frame[1]:
                ultCauseFrame = stack[i + 1]
                break

    if causeFrame is None:
        log.debug('{}: Unknown cause'.format(desc))
    else:
        causeFile = causeFrame[1]
        causeLine = causeFrame[2]
        causeFunc = causeFrame[3]
        causeSrc  = causeFrame[4][causeFrame[5]]

        line = '{}: Caused by {} ({}:{}:{})'.format(
            desc,
            causeFunc,
            op.basename(causeFile),
            causeLine,
            causeSrc.strip())

        if ultCauseFrame is not None:

            causeFile = ultCauseFrame[1]
            causeLine = ultCauseFrame[2]
            causeFunc = ultCauseFrame[3]
            causeSrc  = ultCauseFrame[4][ultCauseFrame[5]]
            line = '{} (ultimately caused by {} ({}:{}:{})'.format(
                line,
                causeFunc,
                op.basename(causeFile),
                causeLine,
                causeSrc.strip())

        log.debug(line)



#!/usr/bin/env python
#
# test_propertyvalue.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import fsleyes_props.properties_value as properties_value

import logging

logging.basicConfig()
logging.getLogger('fsleyes_props').setLevel(logging.DEBUG)

class Context():
    pass


def test_listener():

    ctx    = Context()
    pv     = properties_value.PropertyValue(ctx)
    called = {}

    # TODO Test args passed to listener

    def l1(*a):
        called['l1'] = called.get('l1', 0) + 1

    def l2(*a):
        called['l2'] = called.get('l2', 0) + 1

    def al1(*a):
        called['al1'] = called.get('al1', 0) + 1

    def al2(*a):
        called['al2'] = called.get('al2', 0) + 1

    pv.addListener(         'l1',  l1)
    pv.addListener(         'l2',  l2)
    pv.addAttributeListener('al1', al1)
    pv.addAttributeListener('al2', al2)

    assert pv.hasListener('l1')
    assert pv.hasListener('l2')

    pv.set('New value')

    assert called['l1']  == 1
    assert called['l2']  == 1

    pv.set('New value')

    assert called['l1']  == 1
    assert called['l2']  == 1

    pv.set('New new value')

    assert called['l1']  == 2
    assert called['l2']  == 2

    assert called.get('al1') is None
    assert called.get('al2') is None

    pv.setAttribute('newAtt', 'value')

    assert called['al1'] == 1
    assert called['al2'] == 1

    pv.setAttribute('newAtt', 'value')

    assert called['al1'] == 1
    assert called['al2'] == 1

    pv.setAttribute('newAtt', 'new value')

    assert called['al1'] == 2
    assert called['al2'] == 2



#!/usr/bin/env python
#
# test_widget_bounds.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import wx
import numpy as np

import fsleyes_props               as props
import fsleyes_widgets.rangeslider as rangeslider


from . import (run_with_wx, simtext, simclick, addall, realYield)

import logging
logging.basicConfig()

logging.getLogger('fsleyes_props').setLevel(logging.DEBUG)


def setup_module():
    props.initGUI()


class MyObj(props.HasProperties):

    mybounds = props.Bounds(ndims=1)


def  test_widget_bounds(): run_with_wx(_test_widget_bounds)
def _test_widget_bounds(parent):

    sim = wx.UIActionSimulator()

    obj = MyObj()

    obj.mybounds.xlim = 0, 100
    obj.mybounds.x    = 25, 75

    slispin = props.makeWidget(parent, obj, 'mybounds', showLimits=False)
    sli     = props.makeWidget(parent, obj, 'mybounds', showLimits=False,
                               spin=False, slider=True)

    addall(parent, [slispin, sli])

    slispin = slispin.GetChildren()[0]
    sli     = sli    .GetChildren()[0]

    assert isinstance(slispin, rangeslider.RangeSliderSpinPanel)
    assert isinstance(sli,     rangeslider.RangePanel)

    assert np.all(np.isclose(slispin.GetRange(), (25, 75)))
    assert np.all(np.isclose(sli.GetRange(),     (25, 75)))

    simclick(sim, slispin.lowSlider, pos=(0.5, 0.5))
    val = slispin.GetLow()
    assert np.isclose(obj.mybounds.xlo, val)
    assert np.isclose(sli.GetLow(),     val)

    simclick(sim, slispin.highSlider, pos=(0.9, 0.5))
    val = slispin.GetHigh()
    assert np.isclose(obj.mybounds.xhi, val)
    assert np.isclose(sli.GetHigh(),    val)

    simclick(sim, sli.lowWidget, pos=(0.1, 0.5))
    val = sli.GetLow()
    assert np.isclose(obj.mybounds.xlo, val)
    assert np.isclose(slispin.GetLow(), val)

    simclick(sim, sli.highWidget, pos=(0.6, 0.5))
    val = sli.GetHigh()
    assert np.isclose(obj.mybounds.xhi,  val)
    assert np.isclose(slispin.GetHigh(), val)

    obj.mybounds.x = (66, 99)
    assert np.all(np.isclose(slispin.GetRange(), (66, 99)))
    assert np.all(np.isclose(sli.GetRange(),     (66, 99)))



#!/usr/bin/env python
#
# test_syncable.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import fsleyes_props as props

def test_syncabe_list_link1():  _test_syncabe_list_link(1)
def test_syncabe_list_link2():  _test_syncabe_list_link(2)
def test_syncabe_list_link3():  _test_syncabe_list_link(3)
def test_syncabe_list_link4():  _test_syncabe_list_link(4)
def test_syncabe_list_link5():  _test_syncabe_list_link(5)
def test_syncabe_list_link6():  _test_syncabe_list_link(6)
def test_syncabe_list_link7():  _test_syncabe_list_link(7)
def test_syncabe_list_link8():  _test_syncabe_list_link(8)
def test_syncabe_list_link9():  _test_syncabe_list_link(9)
def test_syncabe_list_link10(): _test_syncabe_list_link(10)

# List properties whcih are synced together, and also
# where, within each instance, individual list items
# are bound across properties
def _test_syncabe_list_link(nchildren):
    class Thing(props.SyncableHasProperties):
        crange = props.Bounds(ndims=1, clamped=False)
        drange = props.Bounds(ndims=1, clamped=False)
        linklo = props.Boolean(default=True)

        def __init__(self, *args, **kwargs):
            props.SyncableHasProperties.__init__(self, *args, **kwargs)
            parent = self.getParent()

            if parent is not None:
                self.addListener('linklo',
                                 str(id(self)),
                                 self.__linkloChanged)
                if self.linklo:  self.__linkloChanged()

        def __linkloChanged(self, *a):
            self.__updateLink(self.linklo, 0)

        def __updateLink(self, val, idx):

            drangePV = self.drange.getPropertyValueList()[idx]
            crangePV = self.crange.getPropertyValueList()[idx]

            if props.propValsAreBound(drangePV, crangePV) == val:
                return

            props.bindPropVals(drangePV,
                               crangePV,
                               bindval=True,
                               bindatt=False,
                               unbind=not val)

            if val:
                crangePV.set(drangePV.get())

    parent   = Thing()
    children = [Thing(parent=parent) for i in range(nchildren)]

    parent.crange = [10, 20]
    parent.drange = [5,  10]

    for i in [parent] + children:
        assert i.crange == [5, 20]
        assert i.drange == [5, 10]
